#include <stdio.h>
#include <math.h>

/*
 Program to prove inequality
 P(X_1+X_2 \ge x/2)/P(X_1 \ge x) -1 + delta < 2(x-1) Lambda / ((a+1)(a+x))
 for a \in [0,1], x \in [4,100]
 where P(X_i \ge x) = (a+1)^2/(a+x)^2 for x\ge 1, X_i independent.

 Algorithm recursively splits rectangle into smaller rectangles
 until one obtains a sufficiently good uniform bound over the rectangle.
 Relies on the fact that P(X_1 \ge x) is decreasing in x and increasing in a.
*/

#define delta (1e-10)	// small value, large enough to compensate for floating pt errors
#define verbose 0	// verbose=1 lists successful rectangles with threshold Lambdas

// P(X_1+X_2 >= x/2)
// requires x>=4 and a>=0
// uses closed form of cdf of X_1+X_2 described in appendix
// note that result is increasing in a and decreasing in x
//
double psum(double a,double x){
 double z1,z2;
 z1=log((2*a+x-2)/(2*a+2))*192*(a+1)*(a+1)/(4*a+x);
 z2=8*((4*a+x)*(4*a+x)+6*(a+1)*(x-4))/(2*a+x-2);
 return (z1+z2)*(a+1)*(a+1)/(4*a+x)/(4*a+x)/(4*a+x);
}

// P(X_1 >= x)
// requires x>=1 and a>=0
// note that result is increasing in a and decreasing in x
//
double pg(double a,double x){
 return (a+1)*(a+1)/(a+x)/(a+x);
}

// Calculates minimum possible \Lambda value
// for which crude bounds on [a1,a2]x[x1,x2] succeed
// requires 0 <= a1 <= a2 and 4 <= x1 <= x2
//
double lambda(double a1,double a2,double x1,double x2){
 double z;
 z=psum(a2,x1)/pg(a1,x2)-1+delta;	// upper bound on LHS
 return z*(a2+1)*(a2+x1)/(x1-1)/2;	// worst case Lambda
// note that (a+1)(a+x)/(x-1)/2 increases as x decreases or a increases
}

// Display successful rectangle / progress report
//
void disp(double a1,double a2,double x1,double x2,double lam){
 static double a=0;
 static int c=10000;
 if(verbose){
  printf("On [%f,%f]x[%f,%f] bound=%f\n",a1,a2,x1,x2,lam);
 }
 else{
  a+=(a2-a1)*(x2-x1);			// area covered
  if((--c)==0||a>95.99){printf("Area %.8f done\n",a);c=10000;} // progress report
 }
}

// Test rectangle
//
int test(double a1,double a2,double x1,double x2,double lam){
 double am,xm,l1,l2,l3,l4;
 if(a2-a1<delta||x2-x1<delta)return 0;	// rectangle too small
 am=(a1+a2)/2;xm=(x1+x2)/2;		// midpoints
 l1=lambda(a1,am,x1,x2);		// split a range in two
 if(l1<lam)disp(a1,am,x1,x2,l1);	// first half successful
 l2=lambda(am,a2,x1,x2);
 if(l2<lam)disp(am,a2,x1,x2,l2);	// second half successful
 if(l1<lam&&l2<lam)return 1;		// both OK, then done
 if(l1<lam)return test(am,a2,x1,x2,lam); // recusively do half that failed 
 if(l2<lam)return test(a1,am,x1,x2,lam);
 if(l2>l1)l1=l2;			// l1 = worst case lambda
 l3=lambda(a1,a2,x1,xm);		// now try splitting x range
 if(l3<lam)disp(a1,a2,x1,xm,l3);
 l4=lambda(a1,a2,xm,x2);
 if(l4<lam)disp(a1,a2,xm,x2,l2);
 if(l3<lam&&l4<lam)return 1;
 if(l3<lam)return test(a1,a2,xm,x2,lam);
 if(l4<lam)return test(a1,a2,x1,xm,lam);
 if(l4>l3)l3=l4;			// l3=worst case lambda
 if(l1<l3){				// splitting a was better
  return test(a1,am,x1,x2,lam) && test(am,a2,x1,x2,lam);
 }
 else{					// splitting x was better
  return test(a1,a2,x1,xm,lam) && test(a1,a2,xm,x2,lam);
 }
}

// Test with fixed Lambda
// optimum Lambda seems to be about 13.0620707
// with worst case (a,x)=(1,9.313068)
// we use a slightly higher Lambda rounded up to 6 decimal places
//
main(){
 double a,x,lam;
 lam=13.062071-delta;			// just less than 13.062071
 if(test(0,1,4,100,lam))printf("\nSucceeded");
 else printf("\nFailed");
 printf(" with lambda=%f\n",lam);
}
