#include <stdio.h>
#include <math.h>

/*
 This program proves certain bounds associated to the coalescence
 process defined in "Coalescence on the real line"

 The aim is to show stochastic domination of an Exp(\Lambda) distribution by
 sums of a geometricaly distributed number of independent uniform distributions.
 The exact requirements are listed in the paper.

 The value of \Lambda can be taken to be 13.062071.
*/

#define lam 13.062071	// upper bound on optimal Lambda
#define EPS (1e-12)	// to compensate for floating point rounding
#define verbose	0	// print x_i if non-zero	

// Calculation of the Cumulative Normal Distribution ---------------------------
/*
 The following are implementations of the error function and (more importantly)
 the complemented error function. They are accurate to almost the full 52 bits
 of double precision floating point numbers.
*/

#define PI 3.141592653589794

// Error function
double erf(double x);
double erfc(double x){
 int n,h;
 double e;
 if(x<2.0)return 1-erf(x);
 h=(int)(22.0/log(x));
 e=sqrt(h/2+x*x/4+0.31)-x/2;
 for(n=h;n;n--)e=n/(x+e)/2;
 return exp(-x*x)/sqrt(PI)/(x+e);
}
double erf(double x){
 double t,s,e;
 int n;
 if(fabs(x)>2)return (x>0?1:-1)*(1-erfc(fabs(x)));
 t=x/sqrt(PI/4); e=t; s=0; n=0;
 while(fabs(t)>1e-16){n++;t*=-x*x/n;s+=t/(2*n+1);}
 return e+s;
}


// Irwin-Hall Distribution -----------------------------------------------------
/* 
 The following calculates the Irwin-Hall distribution of a sum of n independent
 uniform U[0,1] random variables for n not too large. The cumulative
 distribution function is stored as separate polynomials in terms of (x-k) for
 x in the range [k,k+1]. By expanding in terms of (x-k) we keep the coefficients
 small and avoid significant cancellation errors. Results are accurate to almost
 full 52 bit double precision accuracy.
*/

#define N 100		// max order for Irwin-Hall distribution

double ih[N][N][N];	// P(X<x+k)=\sum ih[n][k][i]*x^i, x\in[0,1]

void ihsetup(){		// setup array ih[][][]
 int n,i,k;
 double s;
 for(k=0;k<N;k++){
  for(i=1;i<=N;i++)ih[0][k][i]=0;
  ih[0][k][0]=1;
 }
 for(n=1;n<N;n++){
  ih[n][0][0]=0;
  for(k=1;k<N;k++){
   s=0;
   for(i=1;i<N;i++)s+=ih[n-1][k-1][i-1]/i;
   ih[n][k][0]=s;
  }
  for(i=1;i<N;i++){
   ih[n][0][i]=ih[n-1][0][i-1]/i;
   for(k=1;k<N;k++){
    ih[n][k][i]=(ih[n-1][k][i-1]-ih[n-1][k-1][i-1])/i;
   }
  }
 }
}

/*
 Evaluate Irwin-Hall cumualtive distribution
 P(n.U[0,1]<=x) or provide lower bound via Berry-Esseen if n is too large.
 (Actually Berry-Esseen error should be O(1/n) for symmetric continuous
 distributions, but I could find no explicit bound in the literature.
 Esscher transformed RVs would give better results, but this seems to result
 in only minor improvements.)  
*/
double ihdist(int n,double x){
 int i,k;
 double s;
 if(n>=N){		// use Berry-Esseen
  s=erfc((n*.5-x)/sqrt(n/6.))/2-0.5751/sqrt(n);
  return s<0?0:s;
 }
 k=(int)floor(x);x-=k;
 if(k<0)return 0;
 if(k>=n)return 1;
 if(n>=N||n<0){printf("Error - parameter n=%d\n",n);exit(1);}
 for(s=ih[n][k][n],i=n-1;i>=0;i--)s=s*x+ih[n][k][i];
 return s;
}

//------------------------------------------------------------------------------

/*
 Calculate -log(P(X>x)) where X is a sum of Geom(p) U[a,b] RVs
 The P(X>x) is bounded below, and log is calculated directly so as to
 avoid floating point underflow.
*/
double sumunifdist(double x,double a,double b,double p){
 double s,t;
 int k;
 for(t=s=1/(1-p),k=(int)floor(x/a);k&&t>0;k--){
  s=p*s+(t=ihdist(k,(b*k-x)/(b-a)));
 }
 return -log(s*(1-p))-log(p)*k;
}

main(){
 double x,y,g,p;
 int c,n;
 ihsetup();

 printf("For 1 to beat U[0,1+gamma]\n");
 printf("Max gamma estimate = %f\n",exp(1.5/lam)-1);

 g=0.1216+EPS;
 x=ceil(log(2)/(1/lam-log(1+g)/1.5)+1); // round up a bit
 printf("Taking gamma=%f x_0=%.0f\n",g,x);
 printf("Check 0<gamma<1/2 : ");
 if(g<0.5&&g>0)printf("OK\n"); else {printf("Error\n");exit(1);}
 printf("Check stochastic domination for x>x_0 : ");
 if(log(1+g)*x/1.5+log(2)<x/lam)printf("OK\n"); else {printf("Error\n");exit(1);}
 printf("Check x_i eventually < 1 : ");
 for(c=0;c<10000;c++){
  x=sumunifdist(x,1.,2.,1/(1+g))*lam;
  if(verbose)printf("%f ",x); else printf(".");
  if(x<1)break;
 }
 if(x<1)printf(" OK\nTest succeeded "); else printf(" Error\nTest failed ");
 printf("with gamma=%f\n",g);

 printf("\nFor U[0,1+gamma] to beat 1\n");
 for(g=6.047,c=0;c<20;c++)g=(lam*log((1+g)/(g-1.25))-2)+g/2;
 printf("Min gamma estimate = %f\n",g);

 g=6.048-EPS;
 p=1-(1.25+EPS)/g;
 x=log((1+g)/(p*g))/(2+g/2);
 x=ceil((2*x+log(2))/(1/lam-x)+1); // round up a bit
 printf("Taking gamma=%f p=%f x_0=%.0f\n",g,p,x);
 printf("Check gamma>5 : ");
 if(g>5)printf("OK\n"); else {printf("Error\n");exit(1);}
 printf("Check 5/4<q*gamma<6/4 : ");
 if(g*(1-p)*4>5&&g*(1-p)*4<6)printf("OK\n"); else {printf("Error\n");exit(1);}
 printf("Check stochastic domination for x>x_0 : ");
 if((x+2)/(2+g/2)*log((1+g)/(p*g))+log(2)<x/lam)printf("OK\n"); else {printf("Error\n");exit(1);}
 printf("Check stochastic domination for x<2 : ");
 if((1-p*g/(1+g))/g<exp(-2/lam)/lam)printf("OK\n"); else {printf("Error\n");exit(1);}
 printf("Check x_i eventually < 2 : ");
 for(c=0;c<10000;c++){
  x=sumunifdist(x+2,2.,2.+g,p*g/(1+g))*lam;
  if(verbose)printf("%f ",x); else printf(".");
  if(x<2)break;
 }
 if(x<2)printf(" OK\nTest succeeded "); else printf(" Error\nTest failed ");
 printf("with gamma=%f\n",g);
}

 