#include <stdio.h>

/*
 Monte-Carlo simulation of coalescence process

 We add intervals one by one and coalesce at each step.
 This ensures that the number of intervals we have at
 each step is typically small.

 Aim is to show that usually we have one large blue
 interval within fraction 0.23 of the total length
 from both endpoints
*/  


#define ITER	1000		// number of trials
#define N0	2000000		// half the total number of intervals 		
#define alpha	0.23		// max fraction from each end 
#define N	1000		// max number of active intervals
#define verbose	0		// lists structure of final colouring
#define SEED	1		// random number generator seed 1-65535
#define EPS	0.08
#define EPS2	0.01

double x[N];			// interval lengths

//-------------------------------------------------------------------------------

/* 
 The following is an implementation of a 16-bit ARC4
 random number generator. This is a very high quality RNG.
 Note that the C rand() function is totally inadequate
 for Monte-Carlo simulations.
*/
  
unsigned int rs[1<<16],ri,rj;	// random number generator status

// random 16-bit word
double rand_word(){
 unsigned int s,t;
 ri=(ri+1)&0xFFFF;
 rj=(rj+rs[ri])&0xFFFF;
 s=rs[ri];rs[ri]=t=rs[rj];rs[rj]=s;
 return rs[(s+t)&0xFFFF];
}

// random number uniform in [0,1] (never =0)
double rand_uniform(){
 unsigned int s,t,i;
 double w;
 for(w=.5,i=0;i<4;i++){			// generates 64 bits of random data
  ri=(ri+1)&0xFFFF;
  rj=(rj+rs[ri])&0xFFFF;
  s=rs[ri];rs[ri]=t=rs[rj];rs[rj]=s;
  w=(rs[(s+t)&0xFFFF]+w)/65536;		// rs[(s+t)]=random word
 }
 return w;				// w=0.{64 random bits}1 in binary
}

// initialize random number generator
// s=seed
void rand_setup(int s){
 int i;
 ri=1; rj=s&0xFFFF;
 for(i=0;i<(1<<16);i++)rs[i]=i;
 for(i=0;i<(1<<16);i++)rand_uniform();
}

//-------------------------------------------------------------------------------

// Coalesce c intervals and check that there is a 
// sufficiently large blue interval in the result 
//
int dynamic(int c){
 double y,l,e;
 int j,n;
 for(n=0;c!=0;c--){			// (n&1)=0:blue, (n&1)=1:red
  y=rand_uniform();
//  x[n]=y<0.05?(n&1):2+(y<0.525)*3;	// discrete distribution on {0,1,2,5}

//  x[n]=y<0.05?(n&1)*0.999:1;
//  x[n]+=0.001*rand_uniform();		// plus small uniform to break ties

  x[n]=1+EPS2*y;if(x[n]<1+EPS*EPS2&&(n&1)==0)x[n]-=1;


  while(n>1&&x[n-2]>x[n-1]&&x[n]>x[n-1]){ // collapse intervals
   x[n-2]+=x[n-1]+x[n];n-=2;
  }
  if((++n)==N){printf("\nToo many intervals (increase N)\n");exit(1);}
 }
 if(verbose){
  printf("\n");
  for(j=0;j<n;j++)printf("%.1f ",x[j]);
  printf("\n");
 }
 for(l=0,j=0;j<n;j++)l+=x[j];		// total length
 for(e=0,j=0;j<n&&e<alpha*l;j++)e+=x[j]; // skip leftmost small intervals
 if((j&1)==0)return 0;			// found red - no good
 for(e=0;j<n;j++)e+=x[j];
 if(e>=alpha*l)return 0;		// too much on right
 return 1;
}

main(){
 int i,c;
 printf("Simulation of stochastic dominance counterexample\n");
 printf("U([0,eh]u[1+eh,e]) beats U[1,e]\n");
 printf("with e=%f, h=%f, %d intervals, ",EPS2,EPS,2*N0);
 printf("random seed=%d\n",SEED);
 rand_setup(SEED);
 for(i=c=0;i<ITER;i++){			// number of trials
  if(dynamic(2*N0)){c++;printf("Y");}
  else printf("N");
 }
 printf("\n%d successes out of %d trials\n",c,ITER);
}
